﻿/*
Frog Rescue - Application class
*/

#include <windows.h>
#include <d3d9.h>
#include <stdlib.h>
#include <time.h>
#pragma hdrstop

#include "tlib.h"
#include "LithGUI.h"
#include "LithDX.h"
#include "LithUtil.h"

#include "IStuff.h"

#include "CApplication.h"
#include "CMainFormHandler.h"
#include "CPreLoader.h"
#include "CTitleScreen.h"
#include "CGameController.h"

using namespace tlib;
using namespace LithGUI;
using namespace LithDX;
using namespace LithUtil;



CApplication::CApplication(HINSTANCE aInstance)
{
	srand(time(NULL));

	mGUI = new CGUI(aInstance);
	mDX = new CDX();

	mDevice = NULL;

	mPreLoader = new CPreLoader(this);

	mTitleScreen = NULL;
	mGameController = NULL;

	mGameOver = false;
}



CApplication::~CApplication()
{
	if (mGameController != NULL)
		delete mGameController;
	if (mTitleScreen != NULL)
		delete mTitleScreen;

	if (mDevice != NULL)
		delete mDevice;

	delete mDX;
	delete mGUI;
}



void CApplication::Run()
{
	Initialise();

	if (SetTimer(mMainForm->GetWindow(), 1, 20, NULL) == 0)
		throw TException(L"SetTimer() failed");
	MSG Message;
	while (true)
	{
		int Result = GetMessage(&Message, NULL, 0, 0);
		if (Result == -1)
			throw TException(L"GetMessage() failed");
		if (Result == 0)
			break;
		TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
}



void CApplication::Initialise()
{
	RECT r;
	if (!SetRect(&r, 0, 0, 1024, 768))
		throw TException(L"SetRect() failed");
	if (!AdjustWindowRect(&r, WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_POPUPWINDOW | WS_CAPTION,
			false))
		throw TException(L"AdjustWindowRect() failed");
	mMainForm = mGUI->CreateForm(new CMainFormHandler(this), SL("Frog Rescue"), r.right -
			r.left, r.bottom - r.top);
	mMainForm->SetVisible(true);

	mDevice = mDX->CreateDevice(dmWindow, mMainForm->GetWindow(), 1024, 768);
	mDevice->SetBackgroundColour(D3DCOLOR_XRGB(255, 255, 254));
	mDevice->GetDevice()->SetRenderState(D3DRS_LIGHTING, false);
	mDevice->GetDevice()->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
	mDevice->GetDevice()->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
	mDevice->GetDevice()->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
	mDevice->GetDevice()->SetRenderState(D3DRS_ZENABLE, true);
	mDevice->GetDevice()->SetRenderState(D3DRS_ZWRITEENABLE, true);

	ShowCursor(FALSE);
	//SetCapture(mMainForm->GetWindow());

	// load high-score
	try
	{
		CFile *File = new CFile(SL("Highscore.dat"), omRead);
		mFrogsSavedHigh = File->ReadInt();
		mFrogsSplattedHigh = File->ReadInt();
		delete File;
		mHasHighscore = true;
	}
	catch (TException& e)
	{
		mHasHighscore = false;
	}

	mGUI->CreateThread(mPreLoader);
}



void CApplication::OnDestroy()
{
	//ReleaseCapture();
	::PlaySound(NULL, NULL, 0);
	PostQuitMessage(0);
	if (KillTimer(mMainForm->GetWindow(), 1) == 0)
		throw TException(L"KillTimer() failed");
}



void CApplication::OnTimer()
{
	if (mGameController != NULL)
	{
		POINT mp;
		GetCursorPos(&mp);
		ScreenToClient(mMainForm->GetWindow(), &mp);
		if (mGameController)
			mGameController->SetMousePos(mp.x, mp.y);

		if (GetAsyncKeyState('A'))
			mGameController->CamGoesUp();
		mGameController->Update();
	}
	else
	{
		if (mTitleScreen != NULL)
			mTitleScreen->Update();
	}

	mDevice->BeginFrame();
	if (mGameController != NULL)
		mGameController->Render(mDevice);
	else
	{
		if (mTitleScreen != NULL)
			mTitleScreen->Render(mDevice);
	}
	mDevice->EndFrame();

	if (mGameOver)
	{
		delete mGameController;
		mGameController = NULL;
		mGameOver = false;
	}
}



void CApplication::OnKeyDown(int aKey)
{
	switch (aKey)
	{
	case VK_ESCAPE:
		if (!PostMessage(mMainForm->GetWindow(), WM_CLOSE, 0, 0))
			throw TException(L"PostMessage() failed");
		break;
	case 13:
		if (mGameController == NULL && mPreLoader == NULL)
		{
			mGameController = new CGameController(this);
		}
		break;
	}
}



void CApplication::OnKeyUp(int aKey)
{
}



void CApplication::OnChar(int aChar)
{
}



void CApplication::OnMouseMove(int aX, int aY)
{
}



void CApplication::OnMouse(bool aRight, bool aUp, int aX, int aY)
{
	if (mGameController != NULL)
	{
		if (!aRight && !aUp)
			mGameController->PlaceRoadBlock();
	}
}



CTexture* CApplication::GetTexture(TString aTextureName)
{
	if (!mTextures.ContainsKey(aTextureName))
	{
		TString FileName = aTextureName + SL(".bmp");
		CBitmap *Bitmap = new CBitmap(SL("Data\\") + FileName);
		CTexture *Texture = mDevice->CreateTexture(Bitmap->GetXSize(), Bitmap->GetYSize(),
				Bitmap->GetData());
		delete Bitmap;
		mTextures.Add(aTextureName, Texture);
		return Texture;
	}
	return mTextures[aTextureName];
}



CSprite* CApplication::GetSprite(TString aSpriteName)
{
	if (!mSprites.ContainsKey(aSpriteName))
	{
		TString FileName = aSpriteName + SL(".bmp");
		CBitmap *Bitmap = new CBitmap(SL("Data\\") + FileName);
		CSprite *Sprite = mDevice->CreateSprite(Bitmap->GetXSize(), Bitmap->GetYSize(),
				Bitmap->GetData());
		delete Bitmap;
		mSprites.Add(aSpriteName, Sprite);
		return Sprite;
	}
	return mSprites[aSpriteName];
}



INT_PTR CApplication::GetSound(TString aSoundName)
{
	if (!mSounds.ContainsKey(aSoundName))
	{
		TString FileName = aSoundName + SL(".wav");
		CFile *File = new CFile(SL("Data\\") + FileName, omRead);
		unsigned int Size = File->GetSize();
		TByte *Data = new TByte [Size];
		File->Read(Data, Size);
		delete File;
		mSounds.Add(aSoundName, (INT_PTR)Data);
		return (INT_PTR)Data;
	}
	return mSounds[aSoundName];
}



double CApplication::Random()
{
	return (double)rand() / RAND_MAX;
}



int CApplication::Random(int aStart, int aEnd)
{
	if (!(aEnd >= aStart))
		throw TException(L"Invalid random arguments");
	return rand() % (aEnd - aStart + 1) + aStart;
}



void CApplication::OnUser(int aMessage, WPARAM aWParam, LPARAM aLParam)
{
	switch (aMessage)
	{
	case 0:
		switch (aWParam)
		{
		case -1:   // finished pre-loading
			delete mPreLoader;
			mPreLoader = NULL;
			break;
		case 0:   // title screen data loaded
			for (int i = 0; i < 10; i++)
				mDigits[i] = GetSprite(TString::SPrintF(L"Digit%d", i));
			mTitleScreen = new CTitleScreen(this);
			break;
		}
	}
}



void CApplication::PreLoadInfo(int aInfo)
{
	PostMessage(mMainForm->GetWindow(), WM_USER, aInfo, 0);
}



void CApplication::PlaySound(INT_PTR aSound)
{
	::PlaySound((LPCTSTR)aSound, NULL, SND_MEMORY | SND_ASYNC | SND_NODEFAULT);
}



void CApplication::DrawNumber(int aNumber, int aX, int aY)
{
	if (aNumber < 0)
		throw TException(L"Can't draw this number number");
	TString St = TString::SPrintF(L"%d", aNumber);
	int X = aX;
	for (int i = 0; St.HasNext(i); St.Next(i))
	{
		int Digit = St.GetChar(i) - L'0';
		mDevice->DrawSprite(mDigits[Digit], X, aY, 0, 0);
		X += mDigits[Digit]->GetXSize();
	}
}



void CApplication::GameOver(int aFrogsSaved, int aFrogsSplatted)
{
	mGameOver = true;

	if (!mHasHighscore || aFrogsSaved > mFrogsSavedHigh)
	{
		mHasHighscore = true;
		mFrogsSavedHigh = aFrogsSaved;
		mFrogsSplattedHigh = aFrogsSplatted;

		try
		{
			CFile *File = new CFile(SL("Highscore.dat"), omCreate);
			File->WriteInt(mFrogsSavedHigh);
			File->WriteInt(mFrogsSplattedHigh);
			delete File;
		}
		catch (TException& e)
		{
			// probably because file cannot be written to, just ignore it
		}
	}
}



bool CApplication::GetHighscore(int& aFrogsSaved, int& aFrogsSplatted)
{
	if (mHasHighscore)
	{
		aFrogsSaved = mFrogsSavedHigh;
		aFrogsSplatted = mFrogsSplattedHigh;
	}
	return mHasHighscore;
}