/*
THashTable class
28 February 2008

Implements a map with integer keys.
*/

namespace tlib
{

template <typename V> class THashTable
{
public:
	THashTable()
	{
		FCapacity = 8;
		FRehashCapacity = FCapacity * 0.8;
		FData = new TEntry* [FCapacity];
		memset(FData, 0, FCapacity * sizeof(TEntry*));
		FNumItems = 0;
	}
	~THashTable()
	{
		for (unsigned int i = 0; i < FCapacity; i++)
		{
			TEntry *e = FData[i];
			while (e)
			{
				TEntry *Next = e->Next;
				delete e;
				e = Next;
			}
		}
		delete [] FData;
	}

	void Add(int key, V value)
	{
		if (ContainsKey(key))
			throw TException(L"Key already exists");
		Set(key, value);
	}

	void Set(int key, V value)
	{
		unsigned int Index = (unsigned int)key % FCapacity;
		TEntry *LastEntry = NULL;
		for (TEntry* e = FData[Index]; e; e = e->Next)
		{
			LastEntry = e;
			if (e->Key == key)
			{
				e->Value = value;
				return;
			}
		}
		TEntry *Entry = new TEntry(key, value);
		if (!LastEntry)
			FData[Index] = Entry;
		else
			LastEntry->Next = Entry;
		if (FNumItems == FRehashCapacity)
			Rehash();
		FNumItems++;
	}

	void Remove(int key)
	{
		unsigned int Index = (unsigned int)key % FCapacity;
		TEntry *Last = NULL;
		for (TEntry* e = FData[Index]; e; e = e->Next)
		{
			if (e->Key == key)
			{
				if (!Last)
					FData[Index] = e->Next;
				else
					Last->Next = e->Next;
				delete e;
				FNumItems--;
				// TODO: rehash if NumEntries is below a threshold
				return;
			}
			Last = e;
		}
		throw TException(L"Key does not exist");
	}

	V& operator [](int key) const
	{
		unsigned int Index = (unsigned int)key % FCapacity;
		for (TEntry* e = FData[Index]; e; e = e->Next)
		{
			if (e->Key == key)
				return e->Value;
		}
		throw TException(L"Key does not exist");
	}

	bool ContainsKey(int key) const
	{
		unsigned int Index = (unsigned int)key % FCapacity;
		for (TEntry* e = FData[Index]; e; e = e->Next)
		{
			if (e->Key == key)
				return true;
		}
		return false;
	}

	PROPERTY_F(KeyList, TArray<int>)
	PROPERTY_F(ValueList, TArray<V>)

private:
	struct TEntry
	{
		TEntry(int key, V value)
			: Key(key), Value(value), Next(NULL) { }

		int Key;
		V Value;
		TEntry *Next;
	};
	unsigned int FCapacity;
	unsigned int FRehashCapacity;
	TEntry **FData;
	unsigned int FNumItems;



	/**
	Increases the capacity of the table and rebuilds it.
	*/
	void Rehash()
	{
		TEntry **OldData = FData;
		unsigned int OldCapacity = FCapacity;
		FCapacity += FCapacity;
		FData = new TEntry* [FCapacity];
		memset(FData, 0, FCapacity * sizeof(TEntry*));

		for (unsigned int i = 0; i < OldCapacity; i++)
		{
			TEntry *e = OldData[i];
			while (e)
			{
				TEntry *Next = e->Next;
				e->Next = NULL;
				unsigned int Index = (unsigned int)e->Key % FCapacity;
				TEntry *Last = FData[Index];
				if (!Last)
					FData[Index] = e;
				else
				{
					while (Last->Next)
						Last = Last->Next;
					Last->Next = e;
				}

				e = Next;
			}
		}
		delete [] OldData;

		FRehashCapacity = FCapacity * 0.8;
	}



	TArray<int> _GetKeyList()
	{
		TArray<int> KeyList;
		for (unsigned int i = 0; i < FCapacity; i++)
		{
			for (TEntry *e = FData[i]; e; e = e->Next)
				KeyList.Add(e->Key);
		}
		return KeyList;
	}



	TArray<V> _GetValueList()
	{
		TArray<V> ValueList;
		for (unsigned int i = 0; i < FCapacity; i++)
		{
			for (TEntry *e = FData[i]; e; e = e->Next)
				ValueList.Add(e->Value);
		}
		return ValueList;
	}
};

}